#include "okc_types.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void
okc_coords_set_all_zero (coords* k)
{
  k->x = k->y = k->z = k->a = k->b = k->c = 0.0;
}

void
okc_axis_set_all_zero (lbr_axis* l)
{
  l->a1 = l->a2 = l->a3 = l->a4 = l->a5 = l->a6 = l->e1 = 0.0;
}

void
okc_copy_axis (const lbr_axis src, lbr_axis* dest)
{
  memcpy (dest,&src,sizeof (lbr_axis));
}

void
okc_copy_coords (const coords src, coords* dest)
{
  memcpy (dest,&src,sizeof (coords));
}

void 
okc_print_coords (coords k)
{
  printf ("x = %3.3f\ty = %3.3f\t z = %3.3f\n",k.x,k.y,k.z);
  printf ("a = %3.3f\tb = %3.3f\t c = %3.3f\n",k.a,k.b,k.c);
}


void
okc_print_axis (lbr_axis l)
{
  printf ("A1 = %3.3f\t A2 = %3.3f\t A3 = %3.3f\t A4 = %3.3f\n",l.a1, l.a2,
	  l.a3, l.a4);
  printf ("A5 = %3.3f\t A6 = %3.3f\t E1 = %3.3f\n",l.a5, l.a6, l.e1);
}



okc_handle_t*
okc_malloc_init_okc_handle_t (void) 
{
  okc_handle_t* okch;
  int i;

  okch = malloc (sizeof (okc_handle_t));
  if (NULL == okch)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      return  (NULL);
    }

  okch->flags = 0;
  okch->listener = 0;
  okch->mode = -1;

  
  for (i=0; i < OKC_MAX_ROBOTS; i++)
    {
      okch->robot_map[i] = OKC_SLOT_UNASSIGNED;
      okch->robot_names[i] = NULL;
      okch->update_correction_callback[i] = NULL;
      okch->private_to_update_correction_callback[i] = NULL;
    }

  okch->pos_correction = malloc (OKC_MAX_ROBOTS*sizeof (coords));
  if (NULL == okch->pos_correction)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch);
      return (NULL);
    }

  okch->axis_correction = malloc (OKC_MAX_ROBOTS* sizeof (lbr_axis));
  if (NULL == okch->axis_correction)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch);
      return (NULL);
    }

  okch->pos_act = malloc (OKC_MAX_ROBOTS* sizeof (coords));
  if (NULL == okch->pos_act)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch);
      return (NULL);
    }

  okch->pos_des = malloc (OKC_MAX_ROBOTS* sizeof (coords));
  if (NULL == okch->pos_des)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch);
      return (NULL);
    }

  okch->axis_act = malloc (OKC_MAX_ROBOTS* sizeof (lbr_axis));
  if (NULL == okch->axis_act)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch);
      return (NULL);
    }

  okch->axis_des = malloc (OKC_MAX_ROBOTS* sizeof (lbr_axis));
  if (NULL == okch->axis_des)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch);
      return (NULL);
    }

  okch->torques_axis_act = malloc (OKC_MAX_ROBOTS* sizeof (lbr_axis));
  if (NULL == okch->torques_axis_act)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch->axis_des);
      free (okch);
      return (NULL);
    }

  okch->ft_tcp_est = malloc (OKC_MAX_ROBOTS* sizeof (coords));
  if (NULL == okch->ft_tcp_est)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch->axis_des);
      free (okch->torques_axis_act);
      free (okch);
      return (NULL);
    }

  okch->params = malloc (OKC_MAX_ROBOTS* sizeof (double*));

  if (NULL == okch->params)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch->axis_des);
      free (okch->torques_axis_act);
      free (okch->ft_tcp_est);
      free (okch);
      return (NULL);
    }

  for (i = 0; i < OKC_MAX_ROBOTS; i++)
    {
      okch->params[i] = malloc (7*sizeof (double));
      if (NULL == okch->params[i])
	{
	  perror ("okc_malloc_init_okc_handle_t: malloc");
	  free (okch->pos_correction);
	  free (okch->axis_correction);
	  free (okch->pos_act);
	  free (okch->pos_des);
	  free (okch->axis_act);
	  free (okch->axis_des);
	  free (okch->torques_axis_act);
	  free (okch->ft_tcp_est);
	  free (okch->params);
	  free (okch);
	  return (NULL);
	}
    }


  okch->pos_axis_in_lock = malloc (OKC_MAX_ROBOTS* sizeof (pthread_rwlock_t));
  if (NULL == okch->pos_axis_in_lock)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch->axis_des);
      free (okch->torques_axis_act);
      free (okch->ft_tcp_est);
      for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
      free (okch->params);
      free (okch);
      return (NULL);
    }

  okch->rist_lbra_out_lock = malloc (OKC_MAX_ROBOTS*sizeof (pthread_rwlock_t));
  if (NULL == okch->rist_lbra_out_lock)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch->axis_des);
      free (okch->torques_axis_act);
      free (okch->pos_axis_in_lock);
      free (okch->ft_tcp_est);
      for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
      free (okch->params);
      free (okch);
      return (NULL);
    }
  okch->callback_lock = malloc (OKC_MAX_ROBOTS*sizeof (pthread_rwlock_t));
  
  if (NULL == okch->callback_lock)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->rist_lbra_out_lock);
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch->axis_des);
      free (okch->torques_axis_act);
      free (okch->pos_axis_in_lock);
      free (okch->ft_tcp_est);
      for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
      free (okch->params);
      free (okch);
      return (NULL);
    }
  for (i=0; i < OKC_MAX_ROBOTS; i++)
    {
      if (0 != pthread_rwlock_init (&okch->pos_axis_in_lock[i],NULL))
	{
	  perror ("okc_malloc_init_okc_handle_t: pthread_rwlock_init");
	  free (okch->callback_lock);
	  free (okch->pos_correction);
	  free (okch->axis_correction);
	  free (okch->pos_act);
	  free (okch->pos_des);
	  free (okch->axis_act);
	  free (okch->axis_des);
	  free (okch->torques_axis_act);
	  free (okch->ft_tcp_est);
	  for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
	  free (okch->params);
	  free (okch->pos_axis_in_lock);
	  free (okch->rist_lbra_out_lock);
	  free (okch);
	  return (NULL);
	}
      if (0 != pthread_rwlock_init (&okch->rist_lbra_out_lock[i],NULL))
	{
	  perror ("okc_malloc_init_okc_handle_t: pthread_rwlock_init");
	  free (okch->callback_lock);
	  free (okch->pos_correction);
	  free (okch->axis_correction);
	  free (okch->pos_act);
	  free (okch->pos_des);
	  free (okch->axis_act);
	  free (okch->axis_des);
	  free (okch->torques_axis_act);
	  free (okch->ft_tcp_est);
	  for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
	  free (okch->params);
	  free (okch->pos_axis_in_lock);
	  free (okch->rist_lbra_out_lock);
	  free (okch);
	  return (NULL);
	}
      if (0 != pthread_rwlock_init (&okch->callback_lock[i],NULL))
	{
	  perror ("okc_malloc_init_okc_handle_t: pthread_rwlock_init");
	  free (okch->callback_lock);
	  free (okch->pos_correction);
	  free (okch->axis_correction);
	  free (okch->pos_act);
	  free (okch->pos_des);
	  free (okch->axis_act);
	  free (okch->axis_des);
	  free (okch->torques_axis_act);
	  free (okch->ft_tcp_est);
	  for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
	  free (okch->params);
	  free (okch->pos_axis_in_lock);
	  free (okch->rist_lbra_out_lock);
	  free (okch);
	  return (NULL);
	}
    }

  okch->robot_map_names_lock = malloc (sizeof (pthread_rwlock_t));
  if (NULL == okch->robot_map_names_lock)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch->axis_des);
      free (okch->torques_axis_act);
      free (okch->ft_tcp_est);
      for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
      free (okch->params);
      free (okch->pos_axis_in_lock);
      free (okch->rist_lbra_out_lock);
      free (okch->callback_lock);
      free (okch);
      return (NULL);
    }
  
  if (0 != pthread_rwlock_init (okch->robot_map_names_lock, NULL))
    {
      perror ("okc_malloc_init_okc_handle_t: pthread_rwlock_init");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch->axis_des);
      free (okch->torques_axis_act);
      free (okch->ft_tcp_est);
      for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
      free (okch->params);
      free (okch->pos_axis_in_lock);
      free (okch->rist_lbra_out_lock);
      free (okch->robot_map_names_lock);
      free (okch->callback_lock);
      free (okch);
      return (NULL);
    }
  
  okch->comm_thread = malloc (sizeof (pthread_t));

  if (NULL == okch->comm_thread)
    {
      perror ("okc_malloc_init_okc_handle_t: malloc");
      free (okch->pos_correction);
      free (okch->axis_correction);
      free (okch->pos_act);
      free (okch->pos_des);
      free (okch->axis_act);
      free (okch->axis_des);
      free (okch->torques_axis_act);
      free (okch->ft_tcp_est);
      for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
      free (okch->params);
      free (okch->pos_axis_in_lock);
      free (okch->rist_lbra_out_lock);
      free (okch->robot_map_names_lock);
      free (okch->callback_lock);
      free (okch);
      return (NULL);
    }
  
  return okch;
}

okc_handle_t*
okc_destroy_okc_handle_t (okc_handle_t* okch)
{
  int i;

  for (i=0; i < OKC_MAX_ROBOTS; i++)
    {
      if (0 != pthread_rwlock_destroy (&okch->pos_axis_in_lock[i]))
	{
	  perror ("okc_destroy_okc_handle_t: pthread_rwlock_destroy");
	}
      if (0 != pthread_rwlock_destroy (&okch->rist_lbra_out_lock[i]))
	{
	  perror ("okc_destroy_okc_handle_t: pthread_rwlock_destroy");
	}
    }
  if (0 != pthread_rwlock_destroy (okch->robot_map_names_lock))
    {
      perror ("okc_destroy_okc_handle_t: pthread_rwlock_destroy");
    }
  free (okch->pos_correction);
  free (okch->axis_correction);
  free (okch->pos_act);
  free (okch->pos_des);
  free (okch->axis_act);
  free (okch->axis_des);
  free (okch->torques_axis_act);
  free (okch->ft_tcp_est);
  for (i=0;i < OKC_MAX_ROBOTS;free (okch->params[i++]));
  free (okch->params);
  free (okch->pos_axis_in_lock);
  free (okch->rist_lbra_out_lock);
  free (okch->robot_map_names_lock);
  free (okch->comm_thread);
  free (okch);
  
  return (NULL);
}

void
okc_subtract_coords (coords *difference, coords minuend, coords subtrahend)
{
  difference->x = minuend.x - subtrahend.x;
  difference->y = minuend.y - subtrahend.y;
  difference->z = minuend.z - subtrahend.z;
  difference->a = minuend.a - subtrahend.a;
  difference->b = minuend.b - subtrahend.b;
  difference->c = minuend.c - subtrahend.c;
}

void
okc_subtract_axis (lbr_axis *difference, lbr_axis minuend, lbr_axis subtrahend)
{
  difference->a1 = minuend.a1 - subtrahend.a1;
  difference->a2 = minuend.a2 - subtrahend.a2;
  difference->e1 = minuend.e1 - subtrahend.e1;
  difference->a3 = minuend.a3 - subtrahend.a3;
  difference->a4 = minuend.a4 - subtrahend.a4;
  difference->a5 = minuend.a5 - subtrahend.a5;
  difference->a6 = minuend.a6 - subtrahend.a6;
}
